(* small library providing functions used by the X simulator *)

open Graphics ;;

(* The colors for the buttons *)
let color_up   =  (100, 200, 50) ;;      (* dark green *)
let color_down = (230, 230, 230) ;;      (* light gray *)
let color_wait = (150, 230, 230) ;;

(* draw a rectangle *)
let draw_rect x y size_x size_y = 
  moveto x y;
  lineto (x+size_x) y;
  lineto (x+size_x) (y+size_y);
  lineto x (y+size_y);
  lineto x y ;;

(* Draw a button (a rectangle with a label) *)
let draw_button x y size_x size_y title (r, g, b) =
  set_color black;
  draw_rect x y size_x size_y; 
  set_color (rgb r g b);
  fill_rect (x+1) (y+1) (size_x-1) (size_y-1);
  set_color black;
  let (sx, sy) = text_size title in
  moveto (x+(size_x-sx)/2) (y+(size_y-sy)/2);
  draw_string title ;;

let display_button (x,y,sx,sy) t tf = 
  let c = if tf then color_up else color_down in
  draw_button x y sx sy t c ;;

let make_button x y sx sy t = draw_button x y sx sy t color_down ;;

(* output functions *)
let x_out_bool x y sx sy b = 
  draw_button x y sx sy (if b then "true" else "false") 
  (if b then color_up else color_down) ;;  
let x_out_int x y sx sy i = 
  draw_button x y sx sy (string_of_int i) color_up ;;
let x_out_float x y sx sy f =
  draw_button x y sx sy (string_of_float f) color_up ;;
let x_out_string x y sx sy s =
  draw_button x y sx sy s color_up ;;
let x_out_char x y sx sy c =
  draw_button x y sx sy (Char.escaped c) color_up ;;

(* input functions *)
let x_in_int (x,y,sx,sy) i = 
  draw_button x y sx sy (string_of_int i) color_down ;;
let x_in_bool x y sx sy b =
  draw_button x y sx sy (if b then "true" else "false") color_up ;;  
let x_in_float (x,y,sx,sy) f =
  draw_button x y sx sy (string_of_float f) color_down ;;
let x_in_string (x,y,sx,sy) s =
  draw_button x y sx sy s color_down ;;
let x_in_char x y sx sy c =
  draw_button x y sx sy (Char.escaped c) color_down ;;

(* reading the mouse *)
let mouse_status () = 
  let events = [Button_down;Button_up;Key_pressed] in
  let s = wait_next_event events in
  (((mouse_pos ()),s.button),(s.keypressed,s.key)) ;;

let cons_str old_value key = old_value^(Char.escaped key) ;;

let cons_int old_value key =
  try
    int_of_string ((string_of_int old_value)^(Char.escaped key))
  with
    Failure "int_of_string" -> old_value ;;

let cons_float old_value key = 
  try
    float_of_string ((string_of_float old_value)^(Char.escaped key))
  with
    Failure "float_of_string" -> old_value ;;

let ios nval oval =
  try 
    int_of_string nval
  with
    Failure "int_of_string" -> oval ;;
let fos nval oval =
  try 
    float_of_string nval
  with
    Failure "float_of_string" -> oval ;;  

(* Warning: apparently, "read_key" is buggy *)
(* the following code does not work *)
(* let rec read_entry s = 
   let c = read_key () in
   if (Char.code c) =  13 then s else (
   draw_char c;
   read_entry (s^(Char.escaped c)));;*)

(* this one works *)
let rec read_entry s = 
  if key_pressed () then 
    let c = read_key () in
    if (Char.code c) =  13 then 
      s 
    else (
      draw_char c;
      read_entry (s^(Char.escaped c)))
  else read_entry s ;;

(* same remark as above *)
let rec read_char () = 
  if key_pressed () then 
    read_key()
  else 
    read_char();;

let rec entry_clic conv lect p ecr x y size_x size_y old_value =
  draw_button x y size_x size_y "" color_wait; 
  let (sx, sy) = text_size "0" in
  moveto (x+10) (y+(size_y-sy)/2);
  try
    (conv (lect p)) 
  with Failure "int_of_string" -> 
    draw_button x y size_x size_y (ecr old_value) color_down; 
    old_value 
;;
      
(* int entry *)
let int_clic = entry_clic int_of_string read_entry "" string_of_int;;
(* float entry *)
let float_clic = entry_clic float_of_string read_entry "" string_of_float;;
(* string entry *)
let string_clic = entry_clic (fun x -> x) read_entry "" (fun x->x);;
(* char_entry *)
let char_clic = entry_clic (fun x -> x) read_char () Char.escaped;;

let simul_quit () = close_graph () ; exit 0 ;;

(* clearing the output *)
let clear_out (x, y, sx, sy) = 
  set_color white;
  fill_rect x y sx sy ;;


