open Lucy;;
open Modules;;
open Misc;;
open Initialization;;
open Types;;
open Clocks;;
open Ntree;;

exception Simulator_error of string ;;

(* print a list of string using caml convention for the tuple 
   (s1, s2, ... , sn) *)
let string_of_list l =
  let rec sol_aux = function
      [] -> ""
    | t::q -> if q=[] then t else t^", "^(sol_aux q)
  in
  "(" ^ (sol_aux l) ^ ")" 
;;

(* produce a ntree from a lucy type *)
let ntree_of_type lucytyp =
  let rec ntot dir = function
      Tarrow (_, t1, t2) -> 
      	let rec ntot_arrow = function
	    Tarrow (_, t1, t2) -> (ntot_arrow t1)@[t2]
	  | t -> [t]
	in
	C (List.map (ntot true) ((ntot_arrow t1)@[t2])) 
    | Tproduct l -> if dir then L (List.map (ntot false) l)
              	       	   else C (List.map (ntot true) l)
    | Tconstr (GRname s, l) -> F s
    | Tstream t -> ntot dir t
    | _ -> raise (Simulator_error "ntree_of_type: can't use this type")
  in
  ntot true lucytyp
;;

(* produce a list of pair (var_name, index) from a carrier. 
   var_name is the name used by the compiler *)
let rec list_of_pat id patt =
  let rec lop i p =
    match p with
      Hvarcarrier  vn -> (i+1, [(vn, i)])
    | Htuplecarrier pl -> List.fold_left 
	  (fun (ix, l) x -> let (ni, ne) = lop ix x in (ni, ne@l)) (i, []) pl
    | Hwildcardcarrier -> (i, [])
  in
  snd (lop id patt)
;;

(* search the clock of each entry *)
let search_clock cl = 
  let rec sc id l = function
      Hunknown -> raise (Simulator_error "search_clock : unknown clock")
    | Hon (c,e) ->
	let {e_desc=ed;e_loc=el;e_type=et;e_clock=ec;e_abs=ea} = e in
	let name = (match ed with
	  Lvar v -> let {var_cli=vc;var_info=vi} = v in (
	  match vi with
	    Llocal l -> l
	  | _ -> raise (Simulator_error "search_clock") )
	| _ -> raise (Simulator_error "search_clock")) in
	let ic = List.assoc name l 
	in (id+1, [(id, "i_"^(string_of_int ic))])
    | Harrow (cam, c1, c2) -> 
	let nl = (
	  match cam with
	    (Hwildcardcarrier,Hwildcardcarrier) -> l
	  | (carrier1,Hwildcardcarrier) -> l@(list_of_pat id carrier1)
	  | _ -> raise (Simulator_error 
			  "This type of clock is not treated.\n")) in
        let (i1, r1) = sc id l c1 in
	let (i2, r2) = sc i1 nl c2 in
	(i2, r1@r2)
    | Hproduct lc -> 
	List.fold_left 
	  (fun (i1, r1) c -> 
	    let (i2, r2) = sc i1 l c in 
	    (i2, r1@r2)) (id, []) lc
    | _ -> (id+1, [(id, "true")])
  in
  let rec find_clock = function
      Harrow (h, c1, c2) -> (
	match c2 with
	  Harrow _ -> let (c, l) = find_clock c2 in (Harrow (h,c1,c),l)
	| _ -> ( match h with
	    (Hwildcardcarrier,Hwildcardcarrier) -> (c1, [])
	  | (carrier1,Hwildcardcarrier) -> (c1, list_of_pat 1 carrier1)
	  | _ -> raise (Simulator_error 
			  "This type of clock is not treated.\n")))
    | _ -> raise (Simulator_error "make_sim_function")
  in
  let (ic, il) = find_clock cl in
  snd (sc 1 il ic)
;;

let gen_input oc n ty cloc =
  let l_cl = search_clock cloc in
  let rec gen_let_in typ indice = 
    Printf.fprintf oc "\tlet i_%s = s_input_%s %s in\n"
      (string_of_int indice) typ (List.assoc indice l_cl)
  in
  let rec aux i = function
      Tproduct l ->
	let rec map_input nbr = function
	    [] -> ([], [], nbr)
	  | t::q -> try  
              let (e,f,nbr2,nl) = aux nbr t in
	      let (r,s,nbr3) = map_input nbr2 q in
	      (e::r, f::s, nbr3)
          with _ -> raise (Simulator_error "gen_input aux")
	in
	let (s,t,nbr2) = map_input i l in 
	let ls = string_of_list s in
	(ls, ls, nbr2, true)
    | Tconstr (GRname "unit", _) -> 
	Printf.fprintf oc "\tread_line ();\n" ;
	("()", "()", i, false)
    | Tconstr (GRname typ, _) ->
	let s = string_of_int i in  
	gen_let_in typ i ; 
	("i_"^s, "i_"^s, i+1, true)
    | Tstream t -> aux i t
    | Tarrow (_, t1, t2) -> 
        let (s1, s11, i1, nl1) = aux i t1 in
        let (s2, s22, i2, nl2) = aux i1 t2 in
        (s1^" "^s2, "("^s11^", "^s22^")", i2, true)
    | _ -> raise (Simulator_error "gen_input")
  in
  let (rs, fs, ni, nl) = aux 1 ty in
  (rs, fs, nl)
;;

let gen_output oc n ty =
  let rec gen_name i = function 
      Tconstr (GRname _, _) -> ("o_" ^ (string_of_int i), i+1)
    | Tproduct l -> 
	let rec map_output nbr = function
	    [] -> ([], nbr)
	  | t::q -> let (e,nbr2) = gen_name nbr t in
	    let (r,nbr3) = map_output nbr2 q in
	    (e::r, nbr3)
	in
	let (s, nbr2) = map_output i l in
	(string_of_list s, nbr2)
    | Tstream t -> gen_name i t
    | _ -> ("", i)
  in
  let rec gen_out i = function 
      Tconstr (GRname "int", _) -> 
        Printf.fprintf oc "\ts_output_int o_%d;\n" i; i+1
    | Tconstr (GRname "bool", _) ->
        Printf.fprintf oc "\ts_output_bool o_%d;\n" i;i+1
    | Tconstr (GRname "float", _) ->
        Printf.fprintf oc "\ts_output_float o_%d;\n" i;i+1
    | Tconstr (GRname "string", _) ->
        Printf.fprintf oc "\ts_output_string o_%d;\n" i;i+1
    | Tproduct l -> 
	let rec map_output nbr = function
	    [] -> nbr
	  | t::q ->
	      let nbr2 = gen_out nbr t in 
	      if (q <> []) then
		Printf.fprintf oc "\tprint_string \",\";\n";
	      let nbr3 = map_output nbr2 q in
	      nbr3
	in
	Printf.fprintf oc ("\tprint_string \"(\";\n");
	let nbr = map_output i l in
	Printf.fprintf oc ("\tprint_string \")\";\n");
	nbr
    | Tstream t -> gen_out i t
    | _ -> i
  in
  let l = fst (gen_name 1 ty) in  
  Printf.fprintf oc "let %s_output %s =\n" n l;
  Printf.fprintf oc "\tprint_string \"-> \";\n";
  gen_out 1 ty;
  Printf.fprintf oc "\t()\n;;\n\n"; 
  l
;;

let gen_header oc n input_files =
  Printf.fprintf oc "(* Simulation file for the node %s *)\n" n ;
  List.iter (fun input_file -> Printf.fprintf oc "open %s;;\n" 
      (String.capitalize (Filename.basename input_file))) input_files;
;;

(* standard input functions *)
let gen_infun oc i = 
  let rec rec_gen_infun lt = function
      Tconstr (GRname "int", _) -> if not (List.mem "int" lt) then begin
	Printf.fprintf oc "let s_input_int cl =
	  \tprint_string ((\"int ? \") when cl);
			     \tread_int (() when cl) ;;\n\n" ; ("int"::lt) end
      else lt
    | Tconstr (GRname "bool", _) -> if not (List.mem "bool" lt) then begin
	Printf.fprintf oc ("let s_input_bool cl =
			     \tprint_string ((\"bool ? (0/1) \") when cl);
						\t(read_int (() when cl) <> 0) ;;\n\n") ; ("bool"::lt) end
    else lt
    | Tconstr (GRname "float", _) -> 
        if not (List.mem "float" lt) then begin
	  Printf.fprintf oc ("let s_input_float cl =
			       \tprint_string ((\"float ? \") when cl); 
						  \tread_float (() when cl) ;;\n\n") ; ("float"::lt) end
	else lt
    | Tconstr (GRname "string", _) -> 
        if not (List.mem "string" lt) then begin
	  Printf.fprintf oc ("let s_input_string cl =
			       \tprint_string ((\"string ? \") when cl); 
						  \tread_line (() when cl) ;;\n\n") ; ("string"::lt) end
	else lt
    | Tproduct l -> let rec map_infun ltu = (function
	  [] -> ltu 
	| t::q -> let nl = rec_gen_infun ltu t in
	  map_infun nl q )
    in map_infun lt l
    | Tstream t -> rec_gen_infun lt t
    | Tarrow (_, t1, t2) -> 
        let nl = rec_gen_infun lt t1 in rec_gen_infun nl t2
    | _ -> []
  in
  rec_gen_infun [] i
;;
  
(* standard ouput functions *)
let gen_outfun oc i = 
  let rec rec_gen_outfun lt = function
      Tconstr (GRname "int", _) -> if not (List.mem "int" lt) then begin
	Printf.fprintf oc "let s_output_int i =
	  \tprint_string (string_of_int i) ;;\n\n" ; 
	  ("int"::lt) end
      else lt
    | Tconstr (GRname "bool", _) -> if not (List.mem "bool" lt) then begin
	Printf.fprintf oc ("let s_output_bool b =
			     \tif b then print_string (\"true\")
							 \t\telse print_string (\"false\") ;;\n\n") ; ("bool"::lt) end
    else lt
    | Tconstr (GRname "float", _) -> if not (List.mem "float" lt) then 
	begin
	  Printf.fprintf oc ("let s_output_float f =
			       \tprint_string (string_of_float f) ;;\n\n") ;
			       ("float"::lt) end
    else lt
    | Tconstr (GRname "string", _) -> if not (List.mem "bool" lt) then
	begin
	  Printf.fprintf oc ("let s_output_string s =
			       \tprint_string s ;;\n\n") ; ("string"::lt) end
    else lt
    | Tproduct l -> let rec map_outfun ltu = (function
	  [] -> ltu
	| t::q -> let nl = rec_gen_outfun ltu t in
			map_outfun nl q )
    in map_outfun lt l
    | Tstream t -> rec_gen_outfun lt t
    | _ -> []
  in
  rec_gen_outfun [] i 
;;
  
(* produce the "while true" simulation loop (this is caml code) *)
let emit_simulation_code oc n tf =
  Printf.fprintf oc
  "
    let main () = 
    let state = ref Lucy__nil in";
  if (not tf) then Printf.fprintf oc "\tprint_string \"-- Return --\";\n";
  Printf.fprintf oc "
    while true do
      let _,new_state = %s (true,true) (!state,()) in
      state := new_state
    done;;

    main();exit(0);;"
  n
;;

(* Get type and clock of the node to simulate *)
let get_main_node_type_and_clock n =
  try 
    let st = gen_instance_type (find_value_type (GRname n)) in
    let (t_input, t_output) =
      match st with 
	Tarrow (_ , t1, t2) -> 
	  let rec find_type = function
	      (t1, Tarrow (_, u1, u2)) -> 
		find_type (Tarrow (Tstatefunction, t1, u1), u2)
	    | (i, o) -> (i, o)
	  in find_type (t1, t2)
      | _ -> raise (Simulator_error "What is this type ?") in
    let cl = gen_instance_clock (find_value_clock (GRname n)) st in 
    (t_input, t_output, cl)
  with
    Desc_not_found -> failwith "Undefined node" 
  
;;
     
(* Check that we can simulate the node *)
(* Currently only the primitive types are supported *)
let rec check_simul_type ty =
  match ty with
    Tproduct l -> List.for_all check_simul_type l
  | Tstream t -> check_simul_type t
  | Tarrow (_, t1, t2) -> (check_simul_type t1) && (check_simul_type t2)
  | t -> primitive_type t 
;;
      
(* Simulate a node *)
(* This is the simulation function for the text-mode simulator *)

let simulate input_files n =
  (* the name of the node to be simulated gives *)
  (* the name of the simulation file *)
  (* checks that the name "n" does not belong to the "input_file" list *)
  let name = 
  if List.mem n input_files
  then (Printf.eprintf "Warning: the name %s is already used.\n\
                          The simulation file will be %s_sim.ml\n" n n;
  n^"_sim")
  else n in

  let (t_input, t_output, cl_input) = 
	get_main_node_type_and_clock n in

  if (not (check_simul_type t_input)) 
      or (not (check_simul_type t_output)) then
     raise (Simulator_error "Bad type");
  
  let oc = open_out (name^"__sim.ls") in

  gen_header oc n input_files;
  gen_infun oc t_input;
  gen_outfun oc t_output;
  let output = gen_output oc n t_output in

  Printf.fprintf oc "let main_%s () = \n" n; 
  let (real_input, false_input, nl) = gen_input oc n t_input cl_input in
  Printf.fprintf oc "\tlet %s = %s %s in\n" output n real_input;
  if nl then
	Printf.fprintf oc "\t%s_output %s;\n\tprint_newline (); ()\n;;\n\n" n output
  else
	Printf.fprintf oc "\t%s_output %s; ()\n;; \n\n" n output;
  close_out oc;

  let oc = open_out (name^"__loop.ml") 
  in
  gen_header oc n ((name^"__sim")::input_files);
  emit_simulation_code oc ("main_"^n) nl;
  close_out oc
  ;; 


(*
   the X simulator 
   ---------------
*)

(* The 'in' buttons or entries - one for each primitive type *)

(* a char entry *)
let x_gen_char_in oc i x y fx fy = 
  let tx = fx-x in let ty = fy-y in
  Printf.fprintf oc "
let local bc res =
    let fh = raising_edge bc in
    let fb = falling_edge bc in
    let rec out = 'a' -> merge fh 
                        (char_clic %d %d %d %d ((pre out) when fh))
                        ((pre out) when not fh)
    in
    x_in_char %d %d %d %d (out when (fh or res));
    out;;

let x_%d_entry m_status res =
  let (mp, mc) = button_clic m_status %d %d %d %d in
  let bc = mp & mc in
  reset local bc res every res
" x y tx ty x y tx ty i x y tx ty 
;;

(* functions creating the inputs and outputs from the node type *)
(* generate the call to the printer function *)
let gen_init_in (x, y, fx, fy, s, d) =
  "x_in_"^s^" "^(string_of_int x)^" "^(string_of_int y)^
  " "^(string_of_int (fx-x))^" "^(string_of_int (fy-y))^" "^d^";\n"

let print_in oc (i, x, y, fx, fy) = function
    "bool" -> "display_button ("^(string_of_int x)^","^(string_of_int y)^
  ","^(string_of_int (fx-x))^","^(string_of_int (fy-y))^") \"false\" false;\n"
  | "int"  -> "x_in_int ("^(string_of_int x)^","^(string_of_int y)^
  ","^(string_of_int (fx-x))^","^(string_of_int (fy-y))^") 0;\n"
  | "float"  -> "x_in_float ("^(string_of_int x)^","^(string_of_int y)^
  ","^(string_of_int (fx-x))^","^(string_of_int (fy-y))^") 0.;\n"
  | "char"  -> x_gen_char_in oc i x y fx fy;
      	      gen_init_in (x, y, fx, fy, "char", "'a'")
  | "string"  -> "x_in_string ("^(string_of_int x)^","^(string_of_int y)^
  ","^(string_of_int (fx-x))^","^(string_of_int (fy-y))^") \"\";\n"
  | "unit" -> "()"
  | _ -> raise (Simulator_error "print_in")
;;

let x_gen_input oc l_in tinput =
  let rec gen_button s = function
      [] -> s
    | (i,t,x,y,fx,fy)::q -> 
      	let s2 = print_in oc (i,x,y,fx,fy) t in 
	gen_button (s^s2) q
  in
  let rec gen_param (i, s) = function
      Tconstr (GRname "unit", _) -> (i+1, s^"(() when real_step)")
    | Tconstr (_, _) -> (i+1, s^"(i_"^(string_of_int i)^" when real_step)")
    | Tproduct l -> 
        let rec gpp i s = function
	    [] -> (i, s)
	  | t::q -> let (i1, s1) = gen_param (i, "") t in 
      	       	    let (i2, s2) = gpp i1 "" q in 
      	       	    if s2 = "" then (i2, s1) else (i2, s1^","^s2)
	in
        let (i2, s2) = gpp i "" l in (i2, s^"("^s2^")")
    | Tstream t -> gen_param (i, s) t
    | Tarrow (_, t1, t2) ->
      	let (i1, s1) = gen_param (i, "") t1 in
	gen_param (i1, s1^" ") t2
    | _ -> raise (Simulator_error "x_gen_input")
  in 
  let init_s = gen_button "" l_in in
  let par_s = gen_param (1, "") tinput in
  (init_s, snd (par_s))
;;

(* create the output box and the output variables from the output type *)
let x_gen_ouput oc name l_out t_output =
  let rec gen_param (i, s) = function
      Tconstr (_, _) -> (i+1, s^"o_"^(string_of_int i))
    | Tproduct l -> 
        let rec gpp i s = function
	    [] -> (i, s)
	  | t::q -> let (i1, s1) = gen_param (i, "") t in 
      	       	    let (i2, s2) = gpp i1 "" q in 
      	       	    if s2 = "" then (i2, s1) else (i2, s1^","^s2)
	in
        let (i2, s2) = gpp i "" l in (i2, s^"("^s2^")")
    | Tstream t -> gen_param (i, s) t
    | Tarrow (_, t1, t2) ->
      	let (i1, s1) = gen_param (i, "") t1 in
	gen_param (i1, s1^" ") t2
    | _ -> raise (Simulator_error "x_gen_output")  
  in 
  let rec gen_out = function
      [] -> ()
    | (i,t,x,y,fx,fy)::q -> Printf.fprintf oc
        "\tx_out_%s %d %d %d %d o_%d ;\n" t x y (fx-x) (fy-y) i ;
	gen_out q
  in
  let out = snd (gen_param (1, "") t_output) in
  Printf.fprintf oc "\nlet %s_output %s =\n" name out;
  gen_out l_out;
  Printf.fprintf oc "\t()\n";
  out
;;

(* the simulation function, 
   all the input have the same clock, 
   we reclock them and call the node we wish to simulate *)
let make_sim_function n name cl = 
  let var_name = function i -> "x_"^(string_of_int i)
  in  
  let rec read_clock id l = function
      Hbase -> let r = var_name id in (id+1, r, r)
    | Hunknown -> raise (Simulator_error "Unknown clock")
    | Hvar x -> let r = var_name id in (id+1, r, r)
    | Hon (c,e) -> 
        let (ni, s, sc) = read_clock id l c in
	let {e_desc=ed;e_loc=el;e_type=et;e_clock=ec;e_abs=ea} = e in
	let name = (match ed with
	  Lvar v -> let {var_cli=vc;var_info=vi} = v in
	  begin
	    match vi with
	      Llocal l -> l
	    | _ -> raise (Simulator_error "make_sim_function2")
	  end
	| _ -> raise (Simulator_error "make_sim_function")) in
	let ic = List.assoc name l 
	in
	(ni, s, "("^sc^" when "^(var_name ic)^")")
    | Harrow (cam, c1, c2) -> 
	let nl = (
	  match cam with
	    (Hwildcardcarrier,Hwildcardcarrier) -> l
	  | (carrier1,Hwildcardcarrier) -> l@(list_of_pat id carrier1)
	  | _ -> raise (Simulator_error 
			  "This type of clock is not treated.\n")) in
        let (i1, s1, sc1) = read_clock id l c1 in
	let (i2, s2, sc2) = read_clock i1 nl c2 in
	(i2, s1^" "^s2, sc1^" "^sc2)
    | Hproduct lc -> 
	let (i, s ,sc) = List.fold_left 
	    (fun (i1, s1, sc1) c -> 
	      let (i2, s2, sc2) = read_clock i1 l c in 
	      (i2, s1^", "^s2, sc1^", "^sc2)) (id, "", "") lc in
	(i, "("^(String.sub s 2 ((String.length s)-2))^")", 
	 "("^(String.sub sc 2 ((String.length sc)-2))^")")
  in
  let rec find_clock = function
      Harrow (h, c1, c2) -> (
	match c2 with
	  Harrow _ -> let (c, l) = find_clock c2 in (Harrow (h,c1,c),l)
	| _ -> ( match h with
	    (Hwildcardcarrier,Hwildcardcarrier) -> (c1,[])
	  | (carrier1,Hwildcardcarrier) -> (c1, list_of_pat 1 carrier1)
	  | _ -> raise (Simulator_error 
			  "This type of clock is not treated.\n")))
    | _ -> raise (Simulator_error "make_sim_function")
  in
  let (in_cl, lv) = find_clock cl in
  let (id,s ,sc) = read_clock 1 lv in_cl in
  "let "^n^"_sim "^s^" = \n  "^name^" "^sc^" ;;" 
;;

(* now, we just need to generate the main node ,
   read the entry,
   step if needed,
   write the output *)

(* Generate the function "main" of the simulation file *)
let x_gen_main oc n sim_f input output lv xo yo xos yos = 
  let rec gen_in = function 
      [] -> ""
    | (j,"unit",_,_,_,_)::r -> ""
    | (j,"bool",x,y,sx,sy)::r -> 
       Printf.fprintf oc "     %s i_%d = flipflop_button (%d,%d,%d,%d) \"true\" \"false\" 
                  m_status reset_input\n"
	   (if j=1 then "let rec" else "and") j x y (sx-x) (sy-y);
        (" or i_"^(string_of_int j))^(gen_in r)
    | (j,typ,x,y,sx,sy)::r ->
	Printf.fprintf oc "    %s i_%d = %s_entry (%d,%d,%d,%d) %s m_status breset\n"
	  (if j=1 then "let rec" else "and") j typ x y (sx-x) (sy-y)
	  (match typ with 
	    "int" -> "0" | "float" -> "0." | "string" -> "\"\"");
	gen_in r
    | (j,_,_,_,_,_)::r -> 
       Printf.fprintf oc "    %s i_%d = x_%d_entry m_status breset\n" 
	   (if j=1 then "let rec" else "and") j j;
       gen_in r 
  in
  Printf.fprintf oc "
%s

let sim_main () =
  let m_status = mouse_status () in

  let %s_aff m_status mode step breset =
    clear_out ((%d,%d,%d,%d) when breset);\n" sim_f n xo yo xos yos;
  let real_step = gen_in (List.rev lv) in 
  Printf.fprintf oc "
    and real_step = merge mode ((step %s) when mode) (step when not mode)
    and reset_input = (false -> pre real_step) or breset in 
    let %s = %s_sim %s in 
    clear_out ((%d,%d,%d,%d) when real_step);
    %s_output %s; ()\n  in\n" 
      real_step output n input xo yo xos yos n output ;

  Printf.fprintf oc "

  let quit = simple_button (%d,5,80,30) \"quit\" m_status in
  let mode = flipflop_button (90,5,80,30) \"autostep\" \"compose\" 
             m_status false in
  let step = simple_button (5,5,80,30) \"step\" m_status in
  let breset = simple_button (175,5,80,30) \"reset\" m_status in
  merge quit (simul_quit (() when quit)) ();

  reset %s_aff m_status mode step breset every breset
;;" (xo+xos-87) n

(* Simulate the node n *)
let x_simulate input_files n =
  let name = (String.capitalize (Filename.basename (List.hd input_files)))^
    "."^n in
  
  let (t_input, t_output, node_cl) = get_main_node_type_and_clock n in
  if (not (check_simul_type t_input)) 
      or (not (check_simul_type t_output)) then
     raise (Simulator_error "Bad type");

  let nt_input = ntree_of_type t_input in
  let nt_output = ntree_of_type t_output in

  (* strange things calculating the size of the graphic window *)
  let li = nbr_lignes nt_input in
  let ci = nbr_col nt_input in
  let lo = nbr_lignes nt_output in
  let co = nbr_col nt_output in

  let size_x = if (ci+co)*65 < 345 then 345/(ci+co) else 65 in
  let (size_y_in, size_y_out) = if li > lo then (35,li*35/lo) 
      	       	       	       	       	   else (lo*35/li,35) in
  let win_x = max ((ci+co)*size_x+10) 345 in
  let win_y = (max (li*size_y_in) (lo*size_y_out))+45 in

  (* simulation file *)
  let oc = open_out (n ^ "__sim.ls") in 
    (* the real thing is here, we put everything in a .ls *)
    gen_header oc n ("xsimul"::"graph"::input_files);  (* the necessary open*)
        
    let l_in = calc_coord 1 5 (win_y-5) size_x size_y_in nt_input in
    let l_out = calc_coord 1 (size_x*ci+10) (win_y-5) 
                                size_x size_y_out nt_output in

    let (sinput1, sinput2) = x_gen_input oc l_in t_input in
    let soutput = x_gen_ouput oc n l_out t_output in

    let sim_f = make_sim_function n name node_cl in
    x_gen_main oc n sim_f sinput2 soutput l_in (ci*size_x+6) 
      	       41 (co*size_x+6) (win_y-45) ;

    close_out oc ;

  (* at least we generate a caml file containing the simulation loop
     and the initialization of the graphic window *) 
  let oc = open_out (n^"__loop.ml")
  in 
  gen_header oc n ((n^"__sim")::input_files);
  Printf.fprintf oc "
open Graphics ;;
open Xsimul ;;

let go () = 
  let state = ref %s.Lucy__nil in
  while true do
    sim_main (true, true) () state
  done ;;

open_graph \" %dx%d+50+50\"; 

make_button 0 %d %d 20 \"input\" ;
make_button %d %d %d 20 \"output\" ;

moveto 0 40 ; lineto %d 40 ;
moveto %d %d ; lineto %d 40 ;

make_button  5 5 80 30 \"step\";
make_button %d 5 80 30 \"quit\";
make_button 90 5 80 30 \"compose\";
make_button 175 5 80 30 \"reset\";

%s

go (); exit (0);; " 
  (String.capitalize (Filename.basename (List.hd input_files))) 
  win_x (win_y+20) win_y (ci*size_x+5) (ci*size_x+5) win_y (win_x - (ci*size_x+5)) win_x (ci*size_x+5) win_y (ci*size_x+5) 
  (win_x-85) (if sinput1="()" then "" else sinput1);
  close_out oc
;;
